#include <tuple>
#include <type_traits>
#include <utility>
#include <variant>

template <typename...>
struct list;

// <push_front>
template <typename, typename...>
struct push_front_impl;

template <template <typename...> typename L, typename... T1s, typename... T2s>
struct push_front_impl<L<T1s...>, T2s...> {
    using type = L<T2s..., T1s...>;
};

template <typename L, typename... Ts>
using push_front = typename push_front_impl<L, Ts...>::type;

static_assert(std::is_same_v<push_front<list<>, int>, list<int>>);
static_assert(std::is_same_v<push_front<std::tuple<char>, int>, std::tuple<int, char>>);
static_assert(std::is_same_v<push_front<std::variant<double>, int, char>, std::variant<int, char, double>>);
// </push_front>

// <push_back>
template <typename, typename...>
struct push_back_impl;

template <template <typename...> typename L, typename... T1s, typename... T2s>
struct push_back_impl<L<T1s...>, T2s...> {
    using type = L<T1s..., T2s...>;
};

template <typename L, typename... Ts>
using push_back = typename push_back_impl<L, Ts...>::type;

static_assert(std::is_same_v<push_back<list<>, int>, list<int>>);
static_assert(std::is_same_v<push_back<std::tuple<int>, char>, std::tuple<int, char>>);
static_assert(std::is_same_v<push_back<std::variant<int>, char, double>, std::variant<int, char, double>>);
// </push_back>

// <pop_front>
template <typename T>
struct identity {
    using type = T;
};

template <typename, typename>
class pop_front_impl;

template <template <typename...> typename L, typename... T1s, std::size_t... indices>
class pop_front_impl<L<T1s...>, std::index_sequence<indices...>> {
    template <typename... T2s>
    static L<typename T2s::type...> types(decltype(reinterpret_cast<void*>(indices))..., T2s*...);

public:
    using type = decltype(types(static_cast<identity<T1s>*>(nullptr)...));
};

template <typename L, std::size_t how_many = 1>
using pop_front = typename pop_front_impl<L, std::make_index_sequence<how_many>>::type;

static_assert(std::is_same_v<pop_front<list<int, char, double>, 0>, list<int, char, double>>);
static_assert(std::is_same_v<pop_front<std::tuple<int, char, double>/*, 1*/>, std::tuple<char, double>>);
static_assert(std::is_same_v<pop_front<std::variant<int, char, double>, 2>, std::variant<double>>);
// </pop_front>

// <pop_back>
template <std::size_t, typename T>
struct indexed_type {
    using type = T;
};

template <typename, typename...>
struct index_types;

template <std::size_t... indices, typename... Ts>
struct index_types<std::index_sequence<indices...>, Ts...>
    : indexed_type<indices, Ts>...
{};

template <typename, std::size_t>
class pop_back_impl;

template <template <typename...> typename L, typename... Ts, std::size_t how_many>
class pop_back_impl<L<Ts...>, how_many> {
    struct indexed_types
        : index_types<std::index_sequence_for<Ts...>, Ts...>
    {};

    template <std::size_t index, typename T>
    static indexed_type<index, T> type_at(indexed_type<index, T>);

    template <std::size_t... indices>
    static L<
        typename decltype(type_at<indices>(indexed_types{}))::type...
    > types(
        std::index_sequence<indices...>
    );

    static_assert(sizeof...(Ts) >= how_many);

public:
    using type = decltype(types(std::make_index_sequence<sizeof...(Ts) - how_many>{}));
};

template <typename L, std::size_t how_many = 1>
using pop_back = typename pop_back_impl<L, how_many>::type;

static_assert(std::is_same_v<pop_back<list<int, char, double>, 0>, list<int, char, double>>);
static_assert(std::is_same_v<pop_back<std::tuple<int, char, double>/*, 1*/>, std::tuple<int, char>>);
static_assert(std::is_same_v<pop_back<std::variant<int, char, double>, 2>, std::variant<int>>);
// </pop_back>

int main(int const, char const* const []) {
    return 0;
}